from dataclasses import dataclass, field

import tierkreis.core.protos.tierkreis.v1alpha1.graph as pg
import tierkreis.core.protos.tierkreis.v1alpha1.signature as ps
from tierkreis.core.function import FunctionDeclaration
from tierkreis.core.tierkreis_graph import Location
from tierkreis.core.types import TypeScheme
from tierkreis.core.utils import map_vals


@dataclass(frozen=True)
class Namespace:
    functions: dict[str, FunctionDeclaration]
    subspaces: dict[str, "Namespace"] = field(default_factory=dict)

    @classmethod
    def from_proto(cls, ps_entry: ps.Namespace) -> "Namespace":
        return cls(
            map_vals(ps_entry.functions, lambda v: v.decl),
            map_vals(ps_entry.subspaces, Namespace.from_proto),
        )

    def to_proto(self) -> ps.Namespace:
        return ps.Namespace(
            functions=map_vals(
                self.functions,
                lambda v: ps.NamespaceItem(
                    decl=v, locations=[pg.Location(location=[])]
                ),
            ),
            subspaces=map_vals(self.subspaces, lambda v: v.to_proto()),
        )

    def get(self, ns: list[str]):
        return self if ns == [] else self.subspaces[ns[0]].get(ns[1:])

    def all_namespaces(self) -> list[list[str]]:
        root: list[list[str]] = [[]]
        return root + [
            [k] + x for k, v in self.subspaces.items() for x in v.all_namespaces()
        ]

    @classmethod
    def empty(cls) -> "Namespace":
        return cls({})


@dataclass(frozen=True)
class Signature:
    root: Namespace
    aliases: dict[str, TypeScheme] = field(default_factory=dict)
    scopes: list[Location] = field(default_factory=list)

    @classmethod
    def from_proto(cls, ps_entry: ps.ListFunctionsResponse) -> "Signature":
        return cls(
            Namespace.from_proto(ps_entry.root),
            {k: TypeScheme.from_proto(v) for k, v in ps_entry.aliases.items()},
            ps_entry.scopes,
        )

    def to_proto(self) -> ps.ListFunctionsResponse:
        return ps.ListFunctionsResponse(
            root=self.root.to_proto(),
            aliases=map_vals(self.aliases, lambda v: v.to_proto()),
            scopes=self.scopes,
        )

    @classmethod
    def empty(cls) -> "Signature":
        return cls(Namespace.empty())



from contextlib import AbstractContextManager, contextmanager, nullcontext
from importlib.util import find_spec
from typing import Iterator

try:
    _TRACING = (
        find_spec("opentelemetry.context") and find_spec("opentelemetry.trace")
    ) is not None
except ModuleNotFoundError:
    _TRACING = False


def span(_tracer, **kwargs) -> AbstractContextManager:
    if _TRACING:
        return _tracer.start_as_current_span(**kwargs)
    else:
        return nullcontext()


@contextmanager
def context_token(context) -> Iterator[None]:
    if context is None:
        yield
        return
    else:
        import opentelemetry.context

        token = opentelemetry.context.attach(context)
        try:
            yield
        finally:
            opentelemetry.context.detach(token)


def get_tracer(_name: str):
    if _TRACING:
        import opentelemetry.trace

        return opentelemetry.trace.get_tracer(_name)
    return None



# ruff: noqa: E402
from typing import Type

import numpy as np
import pytest

from tierkreis.core.types import MapType, StructType, TierkreisType, VecType
from tierkreis.core.values import TierkreisValue

pytket = pytest.importorskip("pytket")
from pytket._tket.circuit import Circuit
from pytket.backends.backendresult import BackendResult
from pytket.circuit import Qubit
from pytket.partition import MeasurementBitMap, MeasurementSetup
from pytket.pauli import Pauli, QubitPauliString
from pytket.utils.outcomearray import OutcomeArray

import tierkreis.common_types as common
from tierkreis.common_types import _qps_from_pytket, _qps_to_pytket
from tierkreis.common_types.circuit import (
    CircuitWrapper,
    ResultWrapper,
    register_pytket_types,
)


@pytest.mark.parametrize(
    "qps",
    [
        QubitPauliString({Qubit(0): Pauli.I, Qubit(1): Pauli.Z, Qubit(2): Pauli.X}),
        QubitPauliString({Qubit(1): Pauli.Y, Qubit(3): Pauli.Z, Qubit(2): Pauli.X}),
        QubitPauliString({Qubit(2): Pauli.X, Qubit(1): Pauli.Y, Qubit(0): Pauli.Y}),
    ],
)
def test_qps(qps: QubitPauliString):
    assert _qps_to_pytket(_qps_from_pytket(qps)) == qps


@pytest.mark.parametrize(
    "mbm",
    [
        MeasurementBitMap(2, [0, 1, 2], True),
        MeasurementBitMap(0, [0], False),
        MeasurementBitMap(0, [0], True),
    ],
)
def test_measurementbitmap(mbm: MeasurementBitMap):
    first_ser = common.MeasurementBitMap.from_pytket(mbm)
    deser = first_ser.to_pytket()
    assert first_ser == common.MeasurementBitMap.from_pytket(deser)


def test_measurementsetup():
    circ = Circuit(1, 1)
    circ.X(0)
    circ.Measure(0, 0)
    ms = MeasurementSetup()
    ms.add_measurement_circuit(circ)

    tensor = dict()
    tensor[Qubit(0)] = Pauli.Z

    mbm = MeasurementBitMap(0, [0], True)
    string = QubitPauliString(tensor)
    ms.add_result_for_term(string, mbm)
    assert ms.verify()

    ms2 = MeasurementSetup()
    circ2 = Circuit(2, 2)
    circ2.X(0)
    circ2.Measure(0, 0)
    circ2.V(1)
    circ2.Measure(1, 1)
    ms2.add_measurement_circuit(circ)
    zi = QubitPauliString()
    zi[Qubit(0)] = Pauli.Z
    mbm2 = MeasurementBitMap(0, [0], False)
    ms2.add_result_for_term(zi, mbm2)

    for m in (ms, ms2):
        first_ser = common.MeasurementSetup.from_pytket(m)
        deser = first_ser.to_pytket()
        assert first_ser == common.MeasurementSetup.from_pytket(deser)


def test_registered_conversions():
    register_pytket_types()
    tk_circ = TierkreisType.from_python(Circuit)
    assert isinstance(tk_circ, StructType)
    assert tk_circ == TierkreisType.from_python(CircuitWrapper)
    tk_res = TierkreisType.from_python(BackendResult)
    assert tk_res == TierkreisType.from_python(ResultWrapper)

    circ = Circuit(2, 2).H(0).CX(0, 1).Measure(0, 0).Measure(1, 1)
    assert TierkreisValue.from_python(circ).to_python(Circuit) == circ

    res = BackendResult(
        shots=OutcomeArray(np.array([[2, 3], [4, 5]], dtype=np.uint8), 15)
    )
    assert TierkreisValue.from_python(res).to_python(BackendResult) == res


@pytest.mark.parametrize(
    "t,expected_tk_ty",
    [
        (common.MeasurementSetup, StructType),
        (common.Distribution, MapType),
        (common.QubitPauliString, VecType),
        (common.UnitID, StructType),
    ],
)
def test_type_conversions(t: Type, expected_tk_ty: type[TierkreisType]):
    tk_ty = TierkreisType.from_python(t)
    assert isinstance(tk_ty, expected_tk_ty)



from __future__ import annotations

import inspect
import json
import typing
import warnings
from abc import ABC, abstractmethod
from dataclasses import Field, dataclass, fields, make_dataclass
from enum import Enum
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    Generic,
    Iterable,
    Mapping,
    Optional,
    Protocol,
    TypeVar,
    cast,
    get_args,
    get_origin,
    get_type_hints,
)
from uuid import UUID

import betterproto

import tierkreis.core.protos.tierkreis.v1alpha1.graph as pg
from tierkreis.core import types as TKType
from tierkreis.core.internal import (
    ClassField,
    FieldExtractionError,
    generic_origin,
    python_struct_fields,
)
from tierkreis.core.opaque_model import (
    OpaqueModel,
)
from tierkreis.core.types import (
    NoneType,
    TierkreisPair,
    TierkreisType,
    TupleLabel,
    UnionTag,
    _extract_literal_type,
    _get_discriminators,
    _get_union_args,
    _is_annotated,
    _is_enum,
    _is_var_length_tuple,
)

T = typing.TypeVar("T")
TKVal1 = TypeVar("TKVal1", bound="TierkreisValue")
TKVal2 = TypeVar("TKVal2", bound="TierkreisValue")


@dataclass
class IncompatiblePyValue(Exception):
    value: Any

    def __str__(self) -> str:
        return (
            f"Could not convert python value to tierkreis value: {self.value}."
            + "Try providing a type annotation to `TierkreisValue.from_python`."
        )


@dataclass
class IncompatibleAnnotatedValue(IncompatiblePyValue):
    type_: typing.Type

    def __str__(self) -> str:
        return (
            f"Could not convert python value: {self.value}, "
            + f"with type annotation: {self.type_}"
        )


@dataclass
class ToPythonFailure(Exception):
    value: TierkreisValue

    def __str__(self) -> str:
        return f"Value {self.value} conversion to python type failed."


class TierkreisValue(ABC):
    _proto_map: Dict[str, typing.Type["TierkreisValue"]] = dict()

    def __init_subclass__(cls, **kwargs) -> None:
        super().__init_subclass__()
        TierkreisValue._proto_map[getattr(cls, "_proto_name")] = cls

    @property
    def _instance_pytype(self) -> typing.Type:
        raise NotImplementedError

    @abstractmethod
    def to_proto(self) -> pg.Value:
        pass

    @abstractmethod
    def viz_str(self) -> str:
        pass

    @classmethod
    def from_proto(cls, value: Any) -> "TierkreisValue":
        """
        Parses a tierkreis type from its protobuf representation.
        """
        name, out_value = betterproto.which_one_of(value, "value")

        try:
            return cls._proto_map[name].from_proto(out_value)
        except KeyError as e:
            raise ValueError(f"Unknown protobuf value type: {name}") from e

    @abstractmethod
    def _to_python_impl(self, type_: typing.Type[T]) -> T | None:
        """Each subclass should implement this method to convert to associated
        python type if possible."""
        ...

    def to_python(self, type_: typing.Type[T]) -> T:
        """
        Converts a tierkreis value to a python value given the desired python type.

        When the specified python type is a type variable, the original
        `TierkreisValue` is returned unchanged. This allows us to write generic
        functions in which values of unknown type are passed on as they are.
        """
        if isinstance(type_, typing.TypeVar):
            return cast(T, self)

        if alternative := TierkreisType._convert_as.get(type_):
            alt_value = self.to_python(alternative)
            return alt_value.from_tierkreis_compatible()
        if type_ is NoneType:
            return cast(T, None)
        if (py_val := self._to_python_impl(type_)) is not None:
            # check if known type to subclass
            return py_val
        if inner_type := _extract_literal_type(type_):
            return self.to_python(inner_type)
        if inner_type := _is_annotated(type_):
            return self.to_python(inner_type)
        if args := _get_union_args(type_):
            # expected type is a union so try converting to each and return the
            # first success.
            for possible in args:
                try:
                    return self.to_python(possible)
                except ToPythonFailure:
                    continue
        raise ToPythonFailure(self)

    @classmethod
    def from_python(
        cls, value: Any, type_: typing.Type | None = None
    ) -> "TierkreisValue":
        "Converts a python value to a tierkreis value. If `type_` is not provided, it is inferred from the value."
        try:
            some_type = type_ or type(value)
            # TODO find workaround for delayed imports
            from tierkreis.core.python import RuntimeGraph
            from tierkreis.core.tierkreis_graph import GraphValue

            if alternative := TierkreisType._convert_as.get(some_type):
                return TierkreisValue.from_python(
                    alternative.new_tierkreis_compatible(value), alternative
                )
            if isinstance(value, OpaqueModel):
                return _opaque_to_struct(value)
            if isinstance(value, TierkreisValue):
                return value
            if isinstance(value, RuntimeGraph):
                return GraphValue(value.graph)
            if value is None:
                return option_none
            if isinstance(value, Enum):
                return VariantValue(value.name, StructValue({}))

            return _val_known_type(some_type, value)
        except IncompatiblePyValue as e:
            if type_ is not None:
                raise IncompatibleAnnotatedValue(value, type_) from e
            raise e

    def try_autopython(self) -> Optional[Any]:
        """Try to automatically convert to a python type without specifying the
        target types.
        Returns None if unsuccesful. Expected to be succesful for "simple" or "leaf"
        types, that are not made up component types.

                :return: Python value if succesful, None if not.
                :rtype: Optional[Any]
        """
        try:
            return self.to_python(self._instance_pytype)
        except (ToPythonFailure, NotImplementedError) as _:
            return None


@dataclass(frozen=True)
class BoolValue(TierkreisValue):
    _proto_name: ClassVar[str] = "boolean"

    value: bool

    @property
    def _instance_pytype(self) -> typing.Type:
        return bool

    def to_proto(self) -> pg.Value:
        return pg.Value(boolean=self.value)

    def _to_python_impl(self, type_: typing.Type[T]) -> T | None:
        if type_ is bool:
            return cast(T, self.value)

    @classmethod
    def from_proto(cls, value: Any) -> "TierkreisValue":
        return cls(value)

    def __str__(self) -> str:
        return f"Bool({self.value})"

    def viz_str(self) -> str:
        return "true" if self.value else "false"


@dataclass(frozen=True)
class StringValue(TierkreisValue):
    _proto_name: ClassVar[str] = "str"
    value: str

    @property
    def _instance_pytype(self) -> typing.Type:
        return str

    def to_proto(self) -> pg.Value:
        return pg.Value(str=self.value)

    def _to_python_impl(self, type_: typing.Type[T]) -> T | None:
        if type_ is str:
            return cast(T, self.value)
        if type_ is UUID:
            return cast(T, self.value)

    @classmethod
    def from_proto(cls, value: Any) -> "TierkreisValue":
        return cls(value)

    def __str__(self) -> str:
        return f"String({repr(self.value)})"

    def viz_str(self) -> str:
        return f'"{self.value}"'


@dataclass(frozen=True)
class IntValue(TierkreisValue):
    _proto_name: ClassVar[str] = "integer"
    value: int

    @property
    def _instance_pytype(self) -> typing.Type:
        return int

    def to_proto(self) -> pg.Value:
        return pg.Value(integer=self.value)

    def _to_python_impl(self, type_: typing.Type[T]) -> T | None:
        if type_ is int:
            return cast(T, self.value)

    @classmethod
    def from_proto(cls, value: Any) -> "TierkreisValue":
        return cls(value)

    def __str__(self) -> str:
        return f"Int({self.value})"

    def viz_str(self) -> str:
        return str(self.value)


@dataclass(frozen=True)
class FloatValue(TierkreisValue):
    _proto_name: ClassVar[str] = "flt"
    value: float

    @property
    def _instance_pytype(self) -> typing.Type:
        return float

    def to_proto(self) -> pg.Value:
        return pg.Value(flt=self.value)

    def _to_python_impl(self, type_: typing.Type[T]) -> T | None:
        if type_ is float:
            return cast(T, self.value)

    @classmethod
    def from_proto(cls, value: Any) -> "TierkreisValue":
        return cls(value)

    def __str__(self) -> str:
        return f"Float({self.value})"

    def viz_str(self) -> str:
        return str(self.value)


@dataclass(frozen=True)
class PairValue(Generic[TKVal1, TKVal2], TierkreisValue):
    _proto_name: ClassVar[str] = "pair"
    first: TKVal1
    second: TKVal2

    @property
    def _instance_pytype(self) -> typing.Type:
        return TierkreisPair[self.first._instance_pytype, self.second._instance_pytype]

    def to_proto(self) -> pg.Value:
        return pg.Value(
            pair=pg.PairValue(
                first=self.first.to_proto(),
                second=self.second.to_proto(),
            )
        )

    def _to_python_impl(self, type_: typing.Type[T]) -> T | None:
        if typing.get_origin(type_) is TierkreisPair:
            type_args = typing.get_args(type_)
            first = self.first.to_python(type_args[0])
            second = self.second.to_python(type_args[1])
            return cast(T, TierkreisPair(first, second))

    @classmethod
    def from_tierkreis_pair(cls, value: TierkreisPair) -> "TierkreisValue":
        return PairValue(
            TierkreisValue.from_python(value.first),
            TierkreisValue.from_python(value.second),
        )

    @classmethod
    def from_proto(cls, value: Any) -> "TierkreisValue":
        pair_value = cast(pg.PairValue, value)
        return PairValue(
            TierkreisValue.from_proto(pair_value.first),
            TierkreisValue.from_proto(pair_value.second),
        )

    def __str__(self) -> str:
        return f"Pair({str(self.first)}, {str(self.second)})"

    def viz_str(self) -> str:
        return f"({self.first.viz_str()}, {self.second.viz_str()})"


@dataclass(frozen=True)
class VecValue(Generic[TKVal1], TierkreisValue):
    _proto_name: ClassVar[str] = "vec"
    values: list[TKVal1]

    @property
    def _instance_pytype(self) -> typing.Type:
        elem_t = (
            TypeVar("E") if len(self.values) == 0 else self.values[0]._instance_pytype
        )
        # Would be better to check all elements are the same.
        return list[elem_t]

    def to_proto(self) -> pg.Value:
        return pg.Value(
            vec=pg.VecValue(vec=[value.to_proto() for value in self.values])
        )

    def _to_python_impl(self, type_: typing.Type[T]) -> T | None:
        if typing.get_origin(type_) in (list, tuple):
            type_args = typing.get_args(type_)
            values = [value.to_python(type_args[0]) for value in self.values]
            return cast(T, values)

    @classmethod
    def from_iterable(cls, value: Iterable) -> "TierkreisValue":
        return VecValue([TierkreisValue.from_python(element) for element in value])

    @classmethod
    def from_proto(cls, value: Any) -> "TierkreisValue":
        vec_value = cast(pg.VecValue, value)
        return VecValue(
            [TierkreisValue.from_proto(element) for element in vec_value.vec]
        )

    def __str__(self) -> str:
        return f"Vec({','.join(map(str, self.values))})"

    def viz_str(self) -> str:
        return f"[{', '.join(val.viz_str() for val in self.values)}]"


@dataclass(frozen=True)
class MapValue(Generic[TKVal1, TKVal2], TierkreisValue):
    _proto_name: ClassVar[str] = "map"
    values: Dict[TKVal1, TKVal2]

    @property
    def _instance_pytype(self) -> typing.Type:
        k = next(iter(self.values.keys()), None)
        key_t = TypeVar("K") if k is None else k._instance_pytype
        v = next(iter(self.values.values()), None)
        val_t = TypeVar("V") if v is None else v._instance_pytype
        # Would be better to check all keys and all values are the same.
        return dict[key_t, val_t]

    def to_proto(self) -> pg.Value:
        return pg.Value(
            map=pg.MapValue(
                pairs=[
                    pg.PairValue(first=key.to_proto(), second=value.to_proto())
                    for key, value in self.values.items()
                ]
            )
        )

    def _to_python_impl(self, type_: typing.Type[T]) -> T | None:
        if typing.get_origin(type_) is dict:
            type_args = typing.get_args(type_)
            values = {
                key.to_python(type_args[0]): value.to_python(type_args[1])
                for key, value in self.values.items()
            }
            return cast(T, values)

    @classmethod
    def from_mapping(cls, value: Mapping) -> "TierkreisValue":
        return MapValue(
            {
                TierkreisValue.from_python(element_key): TierkreisValue.from_python(
                    element_value
                )
                for element_key, element_value in value.items()
            }
        )

    @classmethod
    def from_proto(cls, value: Any) -> "TierkreisValue":
        map_value = cast(pg.MapValue, value)
        entries = {}

        for pair_value in map_value.pairs:
            entry_key = TierkreisValue.from_proto(pair_value.first)
            entry_value = TierkreisValue.from_proto(pair_value.second)
            entries[entry_key] = entry_value

        return MapValue(entries)

    def __str__(self) -> str:
        return f"Map({', '.join(f'{key}: {val}' for key, val in self.values.items())})"

    def viz_str(self) -> str:
        entries = (
            f"{key.viz_str()}: {val.viz_str()}" for key, val in self.values.items()
        )
        return f"{{{', '.join(entries)})}}"


def _opaque_to_struct(opaque: OpaqueModel) -> StructValue:
    return StructValue(
        {opaque.tierkreis_field(): StringValue(opaque.model_dump_json())}
    )


class DataclassInstance(Protocol):
    __dataclass_fields__: ClassVar[dict[str, Field[Any]]]


RowStruct = TypeVar("RowStruct", bound=DataclassInstance)


class StructValue(Generic[RowStruct], TierkreisValue):
    _proto_name: ClassVar[str] = "struct"
    _struct: RowStruct

    def __init__(self, values: dict[str, TierkreisValue]) -> None:
        __AnonStruct = make_dataclass("__AnonStruct", values.keys())
        self._struct = __AnonStruct(**values)

    def __eq__(self, __o: object) -> bool:
        if not isinstance(__o, StructValue):
            return False
        return self.values == __o.values

    @property
    def values(self) -> dict[str, TierkreisValue]:
        return {
            k: getattr(self._struct, k) for k in (f.name for f in fields(self._struct))
        }

    @property
    def _instance_pytype(self) -> typing.Type:
        try:
            # if all the keys are tuple labels, report tuple
            arg_vals = _labeled_dict_to_tuple(self.values)
            instance_pytypes = cast(
                tuple[typing.Type, ...], tuple(v._instance_pytype for _, v in arg_vals)
            )
            return tuple[instance_pytypes]  # type: ignore
        except TupleLabel.UnexpectedLabel:
            return super()._instance_pytype

    def to_proto(self) -> pg.Value:
        return pg.Value(struct=pg.StructValue(map=self.to_proto_dict()))

    def _to_python_impl(self, type_: typing.Type[T]) -> T | None:
        type_origin = typing.get_origin(type_) or type_

        if type_origin is tuple:
            args = typing.get_args(type_)
            arg_vals = _labeled_dict_to_tuple(self.values)
            converted = [v.to_python(args[i]) for i, v in arg_vals]
            return cast(T, tuple(converted))
        vals = self.values
        if inspect.isclass(type_) and issubclass(type_, OpaqueModel):
            ((fieldname, payload),) = vals.items()  # Exactly one entry
            assert fieldname == type_.tierkreis_field()
            assert isinstance(payload, StringValue)
            py_val = json.loads(payload.value)
            return cast(Callable[..., T], type_)(**py_val)
        try:
            class_fields = python_struct_fields(type_)
        except FieldExtractionError as e:
            raise ToPythonFailure(self) from e

        field_values = {}
        non_init_values = {}
        for field in class_fields:
            val = vals[field.name]
            if field.discriminant is not None:
                assert isinstance(val, VariantValue)
                var_types = _get_discriminators(field.type_, field.discriminant)
                assert var_types is not None
                field_type = var_types[val.tag]
                val = val.value
            else:
                field_type = field.type_
            py_val = val.to_python(field_type)
            if field.init:
                # field can be set via __init__ method
                field_values[field.name] = py_val
            else:
                non_init_values[field.name] = py_val
        d_cls = cast(Callable[..., T], type_)(**field_values)
        for k, v in non_init_values.items():
            d_cls.__dict__[k] = v
        return d_cls

    @staticmethod
    def from_proto_dict(values: dict[str, pg.Value]) -> "StructValue":
        return StructValue(
            {name: TierkreisValue.from_proto(value) for name, value in values.items()}
        )

    def to_proto_dict(self) -> dict[str, pg.Value]:
        return {name: value.to_proto() for name, value in self.values.items()}

    @classmethod
    def from_object(cls, value: Any, type_: typing.Type | None = None) -> "StructValue":
        struct_type = type_ or type(value)
        try:
            class_fields = python_struct_fields(struct_type)
        except FieldExtractionError as e:
            raise IncompatiblePyValue(value) from e

        def _extract_value(field: ClassField) -> TierkreisValue:
            field_value = getattr(value, field.name)
            if field_value is None:
                return option_none

            union_args = _get_union_args(field.type_)
            if field.discriminant:
                # try to figure out which variant the field_value is
                tag = getattr(field_value, field.discriminant)
                assert tag is not None
                variant_type = None
                if union_args:

                    def _discriminant_matches(fields: list[ClassField]) -> bool:
                        # check if any field has the discriminant field, the
                        # default of which matches the tag
                        return any(
                            f.name == field.discriminant and f.default == tag
                            for f in fields
                        )

                    # attempt to find the type annotation matching the tag
                    variant_type = next(
                        (
                            a
                            for a in union_args
                            if _discriminant_matches(python_struct_fields(a))
                        ),
                        None,
                    )
                return VariantValue(
                    tag, TierkreisValue.from_python(field_value, variant_type)
                )
            if union_args:
                value_tag = UnionTag.value_type_tag(field_value)
                variant_type = next(
                    (a for a in union_args if UnionTag.type_tag(a) == value_tag), None
                )
                return VariantValue(
                    UnionTag.value_type_tag(field_value),
                    TierkreisValue.from_python(field_value, variant_type),
                )

            return TierkreisValue.from_python(field_value, field.type_)

        return cls({field.name: _extract_value(field) for field in class_fields})

    @classmethod
    def from_tuple(cls, tup: tuple[Any, ...]) -> "StructValue":
        """Generate a `StructValue` from a tuple using tuple member indices as labels."""
        return StructValue(
            {
                TupleLabel.index_label(i): TierkreisValue.from_python(v)
                for i, v in enumerate(tup)
            }
        )

    @classmethod
    def from_proto(cls, value: Any) -> "StructValue":
        struct_value = cast(pg.StructValue, value)
        return StructValue(
            {
                name: TierkreisValue.from_proto(value)
                for name, value in struct_value.map.items()
            }
        )

    def __str__(self) -> str:
        key_vals = (f"{key}: {str(val)}" for key, val in self.values.items())
        return f"Struct{{{', '.join(key_vals)}}}"

    def viz_str(self) -> str:
        key_vals = (f"{key}: {val.viz_str()}" for key, val in self.values.items())

        return f"Struct{{{', '.join(key_vals)}}}"


V = TypeVar("V")


def _labeled_dict_to_tuple(labeled_d: dict[str, V]) -> list[tuple[int, V]]:
    # read tuple indices from dictionary keys
    arg_vals = [
        (TupleLabel.parse_label(label), val) for label, val in labeled_d.items()
    ]
    arg_vals.sort(key=lambda x: x[0])
    return arg_vals


@dataclass(frozen=True)
class TierkreisVariant(typing.Generic[T]):
    """Used to represent a Tierkreis Variant as a native python value"""

    tag: str
    value: T


@dataclass(frozen=True)
class VariantValue(Generic[RowStruct], TierkreisValue):
    _proto_name: ClassVar[str] = "variant"
    tag: str
    value: TierkreisValue

    @property
    def _instance_pytype(self) -> typing.Type:
        if self == option_none:
            return Optional[Any]  # type: ignore[reportReturnType]
        return TierkreisVariant[self.value._instance_pytype]

    def to_proto(self) -> pg.Value:
        return pg.Value(
            variant=pg.VariantValue(tag=self.tag, value=self.value.to_proto())
        )

    def _to_python_impl(self, type_: typing.Type[T]) -> T | None:
        if type_ is NoneType and self.tag == UnionTag.none_type_tag():
            return cast(T, None)
        if typing.get_origin(type_) is TierkreisVariant:
            (type_arg,) = typing.get_args(type_)
            return cast(T, TierkreisVariant(self.tag, self.value.to_python(type_arg)))
        if enum_type := _is_enum(type_):
            if not hasattr(enum_type, self.tag):
                raise ToPythonFailure(self)
            return getattr(enum_type, self.tag)
        if args := _get_union_args(type_):
            try:
                tag_type_ = UnionTag.parse_type(self.tag, args)
                return self.value.to_python(tag_type_)
            except UnionTag.UnexpectedTag:
                pass

    def viz_str(self) -> str:
        return f"Variant{{{self.tag}: {self.value.viz_str()}}}"

    @classmethod
    def from_tierkreis_variant(
        cls, value: TierkreisVariant, inner_type: typing.Type | None
    ) -> "TierkreisValue":
        return VariantValue(
            value.tag, TierkreisValue.from_python(value.value, inner_type)
        )

    @classmethod
    def from_proto(cls, value: pg.VariantValue) -> "TierkreisValue":
        return VariantValue(value.tag, TierkreisValue.from_proto(value.value))


option_none: VariantValue = VariantValue(UnionTag.none_type_tag(), StructValue({}))


def _typeddict_to_row(typeddict: dict[str, TierkreisValue]) -> TKType.Row:
    return TKType.Row(
        {key: tkvalue_to_tktype(val) for key, val in get_type_hints(typeddict).items()}
    )


def tkvalue_to_tktype(val_cls: typing.Type[TierkreisValue]) -> TierkreisType:
    if get_origin(val_cls) == VecValue:
        (inner,) = get_args(val_cls)
        return TKType.VecType(tkvalue_to_tktype(inner))
    if get_origin(val_cls) == VariantValue:
        (typeddict,) = get_args(val_cls)
        return TKType.VariantType(_typeddict_to_row(typeddict))

    if get_origin(val_cls) == StructValue:
        (typeddict,) = get_args(val_cls)
        return TKType.StructType(_typeddict_to_row(typeddict))
    if get_origin(val_cls) == PairValue:
        first, second = get_args(val_cls)
        return TKType.PairType(*map(tkvalue_to_tktype, (first, second)))
    if get_origin(val_cls) == MapValue:
        first, second = get_args(val_cls)
        return TKType.MapType(*map(tkvalue_to_tktype, (first, second)))

    if val_cls == IntValue:
        return TKType.IntType()

    if val_cls == BoolValue:
        return TKType.BoolType()

    if val_cls == FloatValue:
        return TKType.FloatType()

    if val_cls == StringValue:
        return TKType.StringType()

    raise ValueError("Cannot convert value class.")


def _val_known_type(type_: typing.Type, value: Any) -> TierkreisValue:
    """Convert a python `value` to a TierkreisValue, given a python type annotation."""
    if type_ is int:
        return IntValue(value)
    if type_ is bool:
        return BoolValue(value)
    if type_ is float:
        if isinstance(value, int):
            # int -> float is an acceptable coercion
            return FloatValue(float(value))
        return FloatValue(value)
    if type_ is str:
        return StringValue(value)

    from tierkreis.core.tierkreis_graph import GraphValue, TierkreisGraph

    if type_ is TierkreisGraph:
        assert isinstance(value, TierkreisGraph)
        return GraphValue(value)

    # types that may contain other types

    type_origin = typing.get_origin(type_) or type_
    type_args = typing.get_args(type_)

    if type_origin is TierkreisVariant:
        try:
            (inner_type,) = type_args
        except ValueError:
            inner_type = None
        return VariantValue.from_tierkreis_variant(value, inner_type)

    if type_origin is tuple:
        if elem_type := _is_var_length_tuple(type_args):
            return VecValue.from_iterable(
                TierkreisValue.from_python(v, elem_type) for v in value
            )
        if type_args:
            return StructValue.from_tuple(
                tuple(
                    TierkreisValue.from_python(v, elem_t)
                    for elem_t, v in zip(type_args, value)
                )
            )
        return StructValue.from_tuple(value)

    if type_origin is list:
        if type_args:
            return VecValue.from_iterable(
                TierkreisValue.from_python(v, type_args[0]) for v in value
            )
        return VecValue.from_iterable(value)
    if type_origin is dict:
        if type_args:
            return MapValue.from_mapping(
                {
                    TierkreisValue.from_python(
                        k, type_args[0]
                    ): TierkreisValue.from_python(v, type_args[1])
                    for k, v in value.items()
                }
            )
        return MapValue.from_mapping(value)
    if type_origin is TierkreisPair:
        if type_args:
            return PairValue(
                TierkreisValue.from_python(value.first, type_args[0]),
                TierkreisValue.from_python(value.second, type_args[1]),
            )
        return PairValue.from_tierkreis_pair(value)
    if inner_type := _extract_literal_type(type_):
        return TierkreisValue.from_python(value, inner_type)
    if inner_type := _is_annotated(type_):
        return TierkreisValue.from_python(value, inner_type)
    if union_args := _get_union_args(type_):
        for possible in union_args:
            # note if some possible type inherits from another possible type, an
            # instance of the child type may be converted as if it were an
            # instance of the parent (possibly losing some fields)
            try:
                with warnings.catch_warnings(record=True) as caught:
                    inner_val = TierkreisValue.from_python(value, possible)
                # Success, so issue warnings (perhaps to next catcher up)
                for w in caught:
                    warnings.warn(w.message, w.category)
                union_tag = UnionTag.type_tag(possible)
                return VariantValue(union_tag, inner_val)
            except IncompatiblePyValue:
                # Suppress (discard) warnings from failed variant
                continue
        raise IncompatiblePyValue(value)

    # base case try treating as class.
    _check_subclass(value, type_)
    return StructValue.from_object(value, type_)


def _check_subclass(value: Any, type_: typing.Type):
    value_type = type(value)
    if isinstance(type_, TypeVar):
        return  # skip check if type erased

    target_type = (
        # value_type could be GenericPydanticModel, list, NonGenericClass.
        # Compare against origin of type_, in case type_ is e.g. GenericPydanticModel[SomeArgs]
        (generic_origin(type_) or type_)
        if generic_origin(value_type) is None
        # value_type is GenericPydanticModel[SomeArgs] (it can never be e.g. list[Args])
        # So compare against exact type_, thus failing if type_ is GenericPydanticModel[DifferentArgs]
        else type_
    )

    if not issubclass(value_type, target_type):
        raise IncompatibleAnnotatedValue(value, type_)

    if value_type != target_type:
        warnings.warn(
            f"Ignoring extra fields in {value_type} beyond those in type {type_}."
        )



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: v1alpha1/jobs.proto
# plugin: python-betterproto
# This file has been @generated

from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Dict,
    List,
    Optional,
)

import betterproto
import grpclib
from betterproto.grpc.grpclib_server import ServiceBase

from .. import (
    graph as _graph__,
    signature as _signature__,
)


if TYPE_CHECKING:
    import grpclib.server
    from betterproto.grpc.grpclib_client import MetadataLike
    from grpclib.metadata import Deadline


@dataclass(eq=False, repr=False)
class StartJobRequest(betterproto.Message):
    graph: "_graph__.Graph" = betterproto.message_field(1)
    inputs: "_graph__.StructValue" = betterproto.message_field(2)
    job_id: str = betterproto.string_field(3)


@dataclass(eq=False, repr=False)
class StartJobResponse(betterproto.Message):
    job_id: str = betterproto.string_field(1, group="result")
    type_errors: "_signature__.TypeErrors" = betterproto.message_field(
        2, group="result"
    )
    runtime_error: str = betterproto.string_field(3, group="result")


@dataclass(eq=False, repr=False)
class RunningJobsRequest(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class RunningJobsResponse(betterproto.Message):
    jobs: List["JobStatus"] = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class StopJobRequest(betterproto.Message):
    id: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class StopJobResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class JobStatusRequest(betterproto.Message):
    id: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class JobStatusResponse(betterproto.Message):
    status: "JobStatus" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class Empty(betterproto.Message):
    """
    This is supposed to be `google.protobuf.Empty` but unfortunately there is
    no support for this in `betterproto` yet.
    """

    pass


@dataclass(eq=False, repr=False)
class JobStatus(betterproto.Message):
    id: str = betterproto.string_field(1)
    success: "Empty" = betterproto.message_field(2, group="status")
    running: "Empty" = betterproto.message_field(3, group="status")
    error: str = betterproto.string_field(4, group="status")


@dataclass(eq=False, repr=False)
class DeleteCompletedRequest(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class DeleteCompletedResponse(betterproto.Message):
    job_ids: List[str] = betterproto.string_field(1)
    """ids of deleted jobs"""


class JobControlStub(betterproto.ServiceStub):
    async def running_jobs(
        self,
        running_jobs_request: "RunningJobsRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "RunningJobsResponse":
        return await self._unary_unary(
            "/tierkreis.v1alpha1.jobs.JobControl/RunningJobs",
            running_jobs_request,
            RunningJobsResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def start_job(
        self,
        start_job_request: "StartJobRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "StartJobResponse":
        return await self._unary_unary(
            "/tierkreis.v1alpha1.jobs.JobControl/StartJob",
            start_job_request,
            StartJobResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def job_status(
        self,
        job_status_request: "JobStatusRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "JobStatusResponse":
        return await self._unary_unary(
            "/tierkreis.v1alpha1.jobs.JobControl/JobStatus",
            job_status_request,
            JobStatusResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def stop_job(
        self,
        stop_job_request: "StopJobRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "StopJobResponse":
        return await self._unary_unary(
            "/tierkreis.v1alpha1.jobs.JobControl/StopJob",
            stop_job_request,
            StopJobResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def delete_completed(
        self,
        delete_completed_request: "DeleteCompletedRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "DeleteCompletedResponse":
        return await self._unary_unary(
            "/tierkreis.v1alpha1.jobs.JobControl/DeleteCompleted",
            delete_completed_request,
            DeleteCompletedResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )


class JobControlBase(ServiceBase):

    async def running_jobs(
        self, running_jobs_request: "RunningJobsRequest"
    ) -> "RunningJobsResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def start_job(
        self, start_job_request: "StartJobRequest"
    ) -> "StartJobResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def job_status(
        self, job_status_request: "JobStatusRequest"
    ) -> "JobStatusResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def stop_job(self, stop_job_request: "StopJobRequest") -> "StopJobResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def delete_completed(
        self, delete_completed_request: "DeleteCompletedRequest"
    ) -> "DeleteCompletedResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def __rpc_running_jobs(
        self, stream: "grpclib.server.Stream[RunningJobsRequest, RunningJobsResponse]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.running_jobs(request)
        await stream.send_message(response)

    async def __rpc_start_job(
        self, stream: "grpclib.server.Stream[StartJobRequest, StartJobResponse]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.start_job(request)
        await stream.send_message(response)

    async def __rpc_job_status(
        self, stream: "grpclib.server.Stream[JobStatusRequest, JobStatusResponse]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.job_status(request)
        await stream.send_message(response)

    async def __rpc_stop_job(
        self, stream: "grpclib.server.Stream[StopJobRequest, StopJobResponse]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.stop_job(request)
        await stream.send_message(response)

    async def __rpc_delete_completed(
        self,
        stream: "grpclib.server.Stream[DeleteCompletedRequest, DeleteCompletedResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.delete_completed(request)
        await stream.send_message(response)

    def __mapping__(self) -> Dict[str, grpclib.const.Handler]:
        return {
            "/tierkreis.v1alpha1.jobs.JobControl/RunningJobs": grpclib.const.Handler(
                self.__rpc_running_jobs,
                grpclib.const.Cardinality.UNARY_UNARY,
                RunningJobsRequest,
                RunningJobsResponse,
            ),
            "/tierkreis.v1alpha1.jobs.JobControl/StartJob": grpclib.const.Handler(
                self.__rpc_start_job,
                grpclib.const.Cardinality.UNARY_UNARY,
                StartJobRequest,
                StartJobResponse,
            ),
            "/tierkreis.v1alpha1.jobs.JobControl/JobStatus": grpclib.const.Handler(
                self.__rpc_job_status,
                grpclib.const.Cardinality.UNARY_UNARY,
                JobStatusRequest,
                JobStatusResponse,
            ),
            "/tierkreis.v1alpha1.jobs.JobControl/StopJob": grpclib.const.Handler(
                self.__rpc_stop_job,
                grpclib.const.Cardinality.UNARY_UNARY,
                StopJobRequest,
                StopJobResponse,
            ),
            "/tierkreis.v1alpha1.jobs.JobControl/DeleteCompleted": grpclib.const.Handler(
                self.__rpc_delete_completed,
                grpclib.const.Cardinality.UNARY_UNARY,
                DeleteCompletedRequest,
                DeleteCompletedResponse,
            ),
        }



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: v1alpha/signature.proto
# plugin: python-betterproto
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Dict,
    List,
    Optional,
)

import betterproto
import grpclib
from betterproto.grpc.grpclib_server import ServiceBase

from .. import graph as _graph__


if TYPE_CHECKING:
    import grpclib.server
    from betterproto.grpc.grpclib_client import MetadataLike
    from grpclib.metadata import Deadline


@dataclass(eq=False, repr=False)
class FunctionDeclaration(betterproto.Message):
    type_scheme: "_graph__.TypeScheme" = betterproto.message_field(2)
    description: str = betterproto.string_field(3)
    input_order: List[str] = betterproto.string_field(4)
    output_order: List[str] = betterproto.string_field(5)


@dataclass(eq=False, repr=False)
class ListFunctionsRequest(betterproto.Message):
    loc: "_graph__.Location" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class ListFunctionsResponse(betterproto.Message):
    root: "Namespace" = betterproto.message_field(1)
    aliases: Dict[str, "_graph__.TypeScheme"] = betterproto.map_field(
        2, betterproto.TYPE_STRING, betterproto.TYPE_MESSAGE
    )
    scopes: List["_graph__.Location"] = betterproto.message_field(3)


@dataclass(eq=False, repr=False)
class InferTypeRequest(betterproto.Message):
    value: "_graph__.Value" = betterproto.message_field(1)
    loc: "_graph__.Location" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class InferTypeResponse(betterproto.Message):
    success: "InferTypeSuccess" = betterproto.message_field(1, group="response")
    error: "TypeErrors" = betterproto.message_field(2, group="response")


@dataclass(eq=False, repr=False)
class InferTypeSuccess(betterproto.Message):
    value: "_graph__.Value" = betterproto.message_field(1)
    type_scheme: "_graph__.TypeScheme" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class GraphWithInputs(betterproto.Message):
    graph: "_graph__.Graph" = betterproto.message_field(1)
    inputs: Optional["_graph__.StructValue"] = betterproto.message_field(
        2, optional=True, group="_inputs"
    )


@dataclass(eq=False, repr=False)
class InferGraphTypesRequest(betterproto.Message):
    gwi: "GraphWithInputs" = betterproto.message_field(1)
    functions: "Namespace" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class InferGraphTypesResponse(betterproto.Message):
    success: "GraphWithInputs" = betterproto.message_field(1, group="response")
    error: "TypeErrors" = betterproto.message_field(2, group="response")


@dataclass(eq=False, repr=False)
class Empty(betterproto.Message):
    """
    This is supposed to be `google.protobuf.Empty` but unfortunately there is
    no support for this in `betterproto` yet.
    """

    pass


@dataclass(eq=False, repr=False)
class GraphLocation(betterproto.Message):
    root: "Empty" = betterproto.message_field(1, group="location")
    vec_index: int = betterproto.uint32_field(2, group="location")
    node_idx: int = betterproto.uint32_field(4, group="location")
    edge: "_graph__.Edge" = betterproto.message_field(5, group="location")
    input: "Empty" = betterproto.message_field(6, group="location")
    output: "Empty" = betterproto.message_field(7, group="location")
    struct_field: str = betterproto.string_field(8, group="location")
    pair_first: "Empty" = betterproto.message_field(9, group="location")
    pair_second: "Empty" = betterproto.message_field(10, group="location")
    map_key: "Empty" = betterproto.message_field(11, group="location")
    map_value: "Empty" = betterproto.message_field(12, group="location")
    input_value: str = betterproto.string_field(13, group="location")


@dataclass(eq=False, repr=False)
class TypeErrors(betterproto.Message):
    errors: List["TierkreisTypeError"] = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class UnifyError(betterproto.Message):
    expected: "_graph__.Type" = betterproto.message_field(1)
    found: "_graph__.Type" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class TypeVarError(betterproto.Message):
    variable: "_graph__.TypeSchemeVar" = betterproto.message_field(1)
    type_scheme: "_graph__.TypeScheme" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class ErrorVariant(betterproto.Message):
    unify: "UnifyError" = betterproto.message_field(1, group="error")
    kind: str = betterproto.string_field(2, group="error")
    unknown_function: "_graph__.FunctionName" = betterproto.message_field(
        3, group="error"
    )
    unknown_type_var: "TypeVarError" = betterproto.message_field(4, group="error")
    bound: str = betterproto.string_field(5, group="error")


@dataclass(eq=False, repr=False)
class TierkreisTypeError(betterproto.Message):
    variant: "ErrorVariant" = betterproto.message_field(1)
    location: List["GraphLocation"] = betterproto.message_field(2)
    """locations go from outermost to innermost in nested const/box graphs"""


@dataclass(eq=False, repr=False)
class NamespaceItem(betterproto.Message):
    decl: "FunctionDeclaration" = betterproto.message_field(1)
    locations: List["_graph__.Location"] = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class Namespace(betterproto.Message):
    functions: Dict[str, "NamespaceItem"] = betterproto.map_field(
        1, betterproto.TYPE_STRING, betterproto.TYPE_MESSAGE
    )
    subspaces: Dict[str, "Namespace"] = betterproto.map_field(
        2, betterproto.TYPE_STRING, betterproto.TYPE_MESSAGE
    )


class SignatureStub(betterproto.ServiceStub):
    async def list_functions(
        self,
        list_functions_request: "ListFunctionsRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "ListFunctionsResponse":
        return await self._unary_unary(
            "/tierkreis.v1alpha.signature.Signature/ListFunctions",
            list_functions_request,
            ListFunctionsResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )


class TypeInferenceStub(betterproto.ServiceStub):
    async def infer_type(
        self,
        infer_type_request: "InferTypeRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "InferTypeResponse":
        return await self._unary_unary(
            "/tierkreis.v1alpha.signature.TypeInference/InferType",
            infer_type_request,
            InferTypeResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )


class SignatureBase(ServiceBase):
    async def list_functions(
        self, list_functions_request: "ListFunctionsRequest"
    ) -> "ListFunctionsResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def __rpc_list_functions(
        self,
        stream: "grpclib.server.Stream[ListFunctionsRequest, ListFunctionsResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.list_functions(request)
        await stream.send_message(response)

    def __mapping__(self) -> Dict[str, grpclib.const.Handler]:
        return {
            "/tierkreis.v1alpha.signature.Signature/ListFunctions": grpclib.const.Handler(
                self.__rpc_list_functions,
                grpclib.const.Cardinality.UNARY_UNARY,
                ListFunctionsRequest,
                ListFunctionsResponse,
            ),
        }


class TypeInferenceBase(ServiceBase):
    async def infer_type(
        self, infer_type_request: "InferTypeRequest"
    ) -> "InferTypeResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def __rpc_infer_type(
        self, stream: "grpclib.server.Stream[InferTypeRequest, InferTypeResponse]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.infer_type(request)
        await stream.send_message(response)

    def __mapping__(self) -> Dict[str, grpclib.const.Handler]:
        return {
            "/tierkreis.v1alpha.signature.TypeInference/InferType": grpclib.const.Handler(
                self.__rpc_infer_type,
                grpclib.const.Cardinality.UNARY_UNARY,
                InferTypeRequest,
                InferTypeResponse,
            ),
        }



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: v1alpha/graph.proto
# plugin: python-betterproto
import builtins
from dataclasses import dataclass
from typing import (
    Dict,
    List,
)

import betterproto


@dataclass(eq=False, repr=False)
class Location(betterproto.Message):
    location: List[str] = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class StructValue(betterproto.Message):
    map: Dict[str, "Value"] = betterproto.map_field(
        1, betterproto.TYPE_STRING, betterproto.TYPE_MESSAGE
    )


@dataclass(eq=False, repr=False)
class PairValue(betterproto.Message):
    first: "Value" = betterproto.message_field(1)
    second: "Value" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class MapValue(betterproto.Message):
    pairs: List["PairValue"] = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class VecValue(betterproto.Message):
    vec: List["Value"] = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class VariantValue(betterproto.Message):
    tag: str = betterproto.string_field(1)
    value: "Value" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class Value(betterproto.Message):
    graph: "Graph" = betterproto.message_field(1, group="value")
    integer: int = betterproto.int64_field(2, group="value")
    boolean: bool = betterproto.bool_field(3, group="value")
    str: builtins.str = betterproto.string_field(9, group="value")
    flt: float = betterproto.double_field(10, group="value")
    pair: "PairValue" = betterproto.message_field(4, group="value")
    vec: "VecValue" = betterproto.message_field(5, group="value")
    map: "MapValue" = betterproto.message_field(7, group="value")
    struct: "StructValue" = betterproto.message_field(8, group="value")
    variant: "VariantValue" = betterproto.message_field(12, group="value")


@dataclass(eq=False, repr=False)
class Output(betterproto.Message):
    edge: "Edge" = betterproto.message_field(1)
    value: "Value" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class OutputStream(betterproto.Message):
    stream: List["Output"] = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class Type(betterproto.Message):
    var: str = betterproto.string_field(1, group="type")
    int: "Empty" = betterproto.message_field(2, group="type")
    bool: "Empty" = betterproto.message_field(3, group="type")
    graph: "GraphType" = betterproto.message_field(4, group="type")
    pair: "PairType" = betterproto.message_field(5, group="type")
    vec: "Type" = betterproto.message_field(6, group="type")
    row: "RowType" = betterproto.message_field(7, group="type")
    map: "PairType" = betterproto.message_field(9, group="type")
    struct: "StructType" = betterproto.message_field(10, group="type")
    str: "Empty" = betterproto.message_field(11, group="type")
    flt: "Empty" = betterproto.message_field(12, group="type")
    variant: "RowType" = betterproto.message_field(14, group="type")
    """Type option = 13;  // REMOVED"""


@dataclass(eq=False, repr=False)
class GraphType(betterproto.Message):
    inputs: "RowType" = betterproto.message_field(1)
    outputs: "RowType" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class PairType(betterproto.Message):
    first: "Type" = betterproto.message_field(1)
    second: "Type" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class RowType(betterproto.Message):
    content: Dict[str, "Type"] = betterproto.map_field(
        1, betterproto.TYPE_STRING, betterproto.TYPE_MESSAGE
    )
    rest: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class StructType(betterproto.Message):
    shape: "RowType" = betterproto.message_field(1)
    name: str = betterproto.string_field(2, group="name_opt")


@dataclass(eq=False, repr=False)
class FunctionName(betterproto.Message):
    namespaces: List[str] = betterproto.string_field(1)
    name: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class BoxNode(betterproto.Message):
    loc: "Location" = betterproto.message_field(1)
    graph: "Graph" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class Node(betterproto.Message):
    input: "Empty" = betterproto.message_field(1, group="node")
    output: "Empty" = betterproto.message_field(2, group="node")
    const: "Value" = betterproto.message_field(3, group="node")
    box: "BoxNode" = betterproto.message_field(4, group="node")
    function: "FunctionName" = betterproto.message_field(5, group="node")
    match: "Empty" = betterproto.message_field(6, group="node")
    tag: str = betterproto.string_field(7, group="node")


@dataclass(eq=False, repr=False)
class Edge(betterproto.Message):
    port_from: str = betterproto.string_field(1)
    port_to: str = betterproto.string_field(2)
    node_from: int = betterproto.uint32_field(3)
    node_to: int = betterproto.uint32_field(4)
    edge_type: "Type" = betterproto.message_field(5)


@dataclass(eq=False, repr=False)
class Graph(betterproto.Message):
    nodes: List["Node"] = betterproto.message_field(1)
    edges: List["Edge"] = betterproto.message_field(2)
    name: str = betterproto.string_field(3)
    input_order: List[str] = betterproto.string_field(4)
    """
    below orderings are optional and used by graph builders if available listed
    ports will be reported and wired up in the listed order and before ports
    that are not listed
    """

    output_order: List[str] = betterproto.string_field(5)


@dataclass(eq=False, repr=False)
class Empty(betterproto.Message):
    """
    This is supposed to be `google.protobuf.Empty` but unfortunately there is
    no support for this in `betterproto` yet.
    """

    pass


@dataclass(eq=False, repr=False)
class TypeScheme(betterproto.Message):
    variables: List["TypeSchemeVar"] = betterproto.message_field(1)
    constraints: List["Constraint"] = betterproto.message_field(2)
    body: "Type" = betterproto.message_field(3)


@dataclass(eq=False, repr=False)
class TypeSchemeVar(betterproto.Message):
    name: str = betterproto.string_field(1)
    kind: "Kind" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class Constraint(betterproto.Message):
    lacks: "LacksConstraint" = betterproto.message_field(1, group="constraint")
    partition: "PartitionConstraint" = betterproto.message_field(2, group="constraint")


@dataclass(eq=False, repr=False)
class PartitionConstraint(betterproto.Message):
    left: "Type" = betterproto.message_field(1)
    right: "Type" = betterproto.message_field(2)
    union: "Type" = betterproto.message_field(3)


@dataclass(eq=False, repr=False)
class LacksConstraint(betterproto.Message):
    row: "Type" = betterproto.message_field(1)
    label: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class Kind(betterproto.Message):
    star: "Empty" = betterproto.message_field(1, group="kind")
    row: "Empty" = betterproto.message_field(2, group="kind")



import inspect
from abc import ABC, abstractmethod
from contextlib import AbstractContextManager
from contextvars import ContextVar, Token
from dataclasses import dataclass
from functools import cached_property
from itertools import count
from types import FrameType, TracebackType
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Generic,
    Iterable,
    Iterator,
    Mapping,
    Optional,
    Type,
    TypeVar,
    Union,
    cast,
    get_args,
    get_origin,
    get_type_hints,
    overload,
)

from yachalk import chalk

from tierkreis.core import Labels
from tierkreis.core.function import FunctionDeclaration, FunctionName
from tierkreis.core.signature import Namespace as SigNamespace
from tierkreis.core.signature import Signature
from tierkreis.core.tierkreis_graph import (
    BoxNode,
    ConstNode,
    FunctionNode,
    IncomingWireType,
    Location,
    MatchNode,
    NodePort,
    NodeRef,
    PortID,
    TagNode,
    TierkreisGraph,
    TierkreisNode,
    to_nodeport,
)
from tierkreis.core.type_errors import TierkreisTypeErrors
from tierkreis.core.type_inference import infer_graph_types
from tierkreis.core.types import TupleLabel, TypeScheme, UnionTag, UnpackRow
from tierkreis.core.utils import graph_from_func, map_vals, rename_ports_graph
from tierkreis.core.values import (
    TierkreisValue,
    TKVal1,
    tkvalue_to_tktype,
)

if TYPE_CHECKING:
    from tierkreis.core.types import GraphType, TierkreisType


# magic for lazily creating nodes (and returning ports thereof)
class PortFunc(ABC):
    @abstractmethod
    def resolve(self) -> NodePort:
        """Return a NodePort, creating nodes as necessary in source graph."""

    @abstractmethod
    def source_graph(self) -> TierkreisGraph:
        """The graph the port belongs to (or will)."""


ValueSource = Union[IncomingWireType, PortFunc]


class StablePortFunc(PortFunc):
    """Extends the contract of PortFunc, as documented for the resolve method"""

    @abstractmethod
    def resolve(self) -> NodePort:
        """As superclass, but guarantees that all calls return the same NodePort."""

    def __getitem__(self, field_name: str) -> "Unpack":
        # Syntactic sugar for unpacking structs
        return Unpack(self, field_name)


def _resolve(s: ValueSource) -> NodePort:
    return s.resolve() if isinstance(s, PortFunc) else to_nodeport(s)


def _source_graph(s: ValueSource) -> TierkreisGraph:
    return (
        s.source_graph() if isinstance(s, PortFunc) else to_nodeport(s).node_ref.graph
    )


def _debug_str() -> str:
    # unroll current stack until out of this file
    frm = cast(FrameType, inspect.currentframe())
    while frm.f_code.co_filename == __file__ and frm.f_back is not None:
        frm = cast(FrameType, frm.f_back)
    try:
        return f"Node added at: {frm.f_code.co_filename}:{frm.f_lineno}"
    finally:
        # as recommended in
        # https://docs.python.org/3/library/inspect.html#the-interpreter-stack
        del frm


__state: ContextVar["GraphBuilder"] = ContextVar("state")


class InvalidContext(Exception):
    """Expression not valid outside GraphBuilder `with` context"""


def current_builder() -> "GraphBuilder":
    try:
        return __state.get()
    except LookupError as e:
        raise InvalidContext() from e


def current_graph() -> TierkreisGraph:
    return current_builder().graph


def _set_state(bg: "GraphBuilder") -> Token["GraphBuilder"]:
    return __state.set(bg)


def _reset_state(tok: Token["GraphBuilder"]):
    return __state.reset(tok)


def _capture_label(idx: int) -> PortID:
    return f"_c{idx}"


@dataclass(frozen=True)
class CaptureOutwards:
    source_graph: TierkreisGraph
    ref: NodeRef
    contained: Optional["CaptureOutwards"] = None

    def capture(self, incoming: ValueSource) -> ValueSource:
        if self.contained is not None:
            incoming = self.contained.capture(incoming)
        new_name = _capture_label(len(self.source_graph.outputs()))
        self.source_graph.set_outputs(**{new_name: _resolve(incoming)})
        return self.ref[new_name]


TGBuilder = TypeVar("TGBuilder", bound="GraphBuilder")


class GraphBuilder(AbstractContextManager):
    graph: TierkreisGraph
    inputs: dict[str, Optional["TierkreisType"]]
    outputs: dict[str, Optional["TierkreisType"]]
    inner_scopes: dict[TierkreisGraph, CaptureOutwards]
    debug_info: dict[NodeRef, str]
    _state_token: Token["GraphBuilder"]

    def __init__(
        self,
        inputs: Optional[Dict[str, Optional["TierkreisType"]]] = None,
        outputs: Optional[Dict[str, Optional["TierkreisType"]]] = None,
        name: str = "",
        debug: bool = True,
    ) -> None:
        self.graph = TierkreisGraph(name)
        self.inputs = inputs or {}
        self.outputs = outputs or {}
        self.inner_scopes = {}
        self.graph.input_order.extend(self.inputs.keys())
        self.debug_info = {self.graph.input: _debug_str()} if debug else {}

    def input(self, name: str) -> NodePort:
        if name in self.inputs:
            return self.graph.input[name]
        raise RuntimeError("Input not declared.")

    def __enter__(self: TGBuilder) -> TGBuilder:
        self._state_token = _set_state(self)
        return self

    def __exit__(
        self,
        __exc_type: Optional[Type[BaseException]],
        __exc_value: Optional[BaseException],
        __traceback: Optional[TracebackType],
    ) -> Optional[bool]:
        _reset_state(self._state_token)
        return None

    def _inner_capture(self, incoming: ValueSource) -> Optional[ValueSource]:
        source_graph = _source_graph(incoming)
        if source_graph is self.graph:
            return incoming
        inner_scope = self.inner_scopes.get(source_graph)
        if inner_scope is not None:
            return inner_scope.capture(incoming)
        return None

    def capture(
        self, incoming: ValueSource, allow_existing: bool = False
    ) -> ValueSource:
        if (ret_val := self._inner_capture(incoming)) is None:
            raise InvalidContext(
                f"ValueSource {incoming} does not belong to current graph."
            )
        return ret_val

    def _add_edges(
        self,
        nr: NodeRef,
        /,
        **incoming_wires: ValueSource,
    ) -> None:
        for tgt_port, vs in incoming_wires.items():
            src = _resolve(self.capture(vs))
            self.graph.add_edge(src, nr[tgt_port])
            if src.node_ref == self.graph.input and (
                inp_typ := self.inputs.get(src.port)
            ):
                self.graph.annotate_input(src.port, inp_typ)

    def add_node_to_graph(
        self,
        _tk_node: TierkreisNode,
        /,
        **incoming_wires: ValueSource,
    ) -> NodeRef:
        nr = self.graph.add_node(_tk_node)
        self._add_edges(nr, **incoming_wires)
        if self.debug_info:
            self.debug_info[nr] = _debug_str()
        return nr

    def set_graph_outputs(self, **incoming_wires: ValueSource) -> None:
        if len(self.graph.outputs()) != 0:
            raise RuntimeError("Outputs set multiple times for same graph.")
        self._add_edges(self.graph.output, **incoming_wires)

    def type_check(self, sig: Signature) -> TierkreisGraph:
        try:
            return infer_graph_types(self.graph, sig)
        except TierkreisTypeErrors as te:
            raise te.with_debug(self.debug_info) from None


class CaptureBuilder(GraphBuilder):
    captured: dict[ValueSource, PortID]

    def __init__(
        self,
        inputs: Optional[Dict[str, Optional["TierkreisType"]]] = None,
        outputs: Optional[Dict[str, Optional["TierkreisType"]]] = None,
        name: str = "",
        debug: Optional[bool] = None,
    ) -> None:
        bool_debug = bool(current_builder().debug_info) if debug is None else debug
        super().__init__(inputs, outputs, name, bool_debug)

        self.captured = {}

    def capture(
        self, incoming: ValueSource, allow_existing: bool = False
    ) -> ValueSource:
        if (valsrc := self._inner_capture(incoming)) is not None:
            return valsrc
        if incoming in self.captured and not allow_existing:
            # This would generate two wires from the same NodePort
            raise ValueError(f"Already captured {incoming} as input to {self.graph}")
        in_name = self.captured.setdefault(incoming, _capture_label(len(self.captured)))
        return self.graph.input[in_name]

    def __exit__(
        self,
        __exc_type: Optional[Type[BaseException]],
        __exc_value: Optional[BaseException],
        __traceback: Optional[TracebackType],
    ) -> Optional[bool]:
        super().__exit__(__exc_type, __exc_value, __traceback)
        if not any((__exc_type, __exc_value, __traceback)):
            current_builder().debug_info.update(self.debug_info)
        return None


def _partial_thunk(
    thunk: Union[GraphBuilder, NodePort], captured: dict[ValueSource, PortID]
) -> NodePort:
    # add the thunk as constant to the current graph and partial up
    bg = current_builder()
    thunk_port: NodePort = (
        Const(thunk.graph) if isinstance(thunk, GraphBuilder) else thunk
    )
    if captured:
        return bg.add_node_to_graph(
            FunctionNode(FunctionName("partial")),
            thunk=thunk_port,
            **{k: v for v, k in captured.items()},
        )[Labels.VALUE]
    return thunk_port


T = TypeVar("T")


def Const(
    val: Any,
    type_: Type | None = None,
) -> NodePort:
    if isinstance(val, LazyGraph):
        val = val.graph
    n = current_builder().add_node_to_graph(
        ConstNode(TierkreisValue.from_python(val, type_))
    )
    return n[Labels.VALUE]


def UnionConst(
    val: Union[
        int,
        float,
        bool,
        str,
        dict,
        TierkreisValue,
        TierkreisGraph,
        tuple,
        list,
        "LazyGraph",
    ],
) -> NodePort:
    """Add a constant as a variant value tagged by its type, to be used
    with `Union` type annotations."""
    union_tag = UnionTag.value_type_tag(val)
    return Tag(union_tag, Const(val))


@dataclass(frozen=True)
class _CallAddNode:
    node: TierkreisNode
    input_order: list[str]
    output_order: list[str]

    def __call__(self, *args: ValueSource, **kwds: ValueSource) -> NodeRef:
        bg = current_builder()
        kwds = _combine_args_with_kwargs(self.input_order, *args, **kwds)

        n = bg.add_node_to_graph(self.node, **kwds)
        return NodeRef(n.idx, n.graph, self.output_order)


class Scope(CaptureBuilder, ABC):
    location: Location

    def __init__(self, location: Union[str, Location] = Location([])):
        if isinstance(location, str):
            self.location = Location(location.split("/"))
        else:
            self.location = location
        super().__init__()

    def __exit__(
        self,
        __exc_type: Optional[Type[BaseException]],
        __exc_value: Optional[BaseException],
        __traceback: Optional[TracebackType],
    ) -> None:
        super().__exit__(__exc_type, __exc_value, __traceback)

        if not any((__exc_type, __exc_value, __traceback)):
            graph = self.graph
            inputs = {v: k for k, v in self.captured.items()}
            node_ref = _build_box(graph, self.location, **inputs)
            for inner_graph, captured in self.inner_scopes.items():
                current_builder().inner_scopes[inner_graph] = CaptureOutwards(
                    graph, node_ref, captured
                )
            current_builder().inner_scopes[graph] = CaptureOutwards(graph, node_ref)


def Tag(tag: str, value: ValueSource) -> NodePort:
    return current_builder().add_node_to_graph(TagNode(tag), value=value)[Labels.VALUE]


OutAnnotation = TypeVar(
    "OutAnnotation",
    bound=Union[UnpackRow, TierkreisValue],
)


class Output(NodeRef, Generic[OutAnnotation]):
    def __init__(self, *args: ValueSource, **kwargs: ValueSource) -> None:
        s = current_builder()

        g = s.graph

        for o in s.outputs:
            g.output_order.append(o)

        if len(args) == 1 and len(kwargs) == 0:
            kwargs = {Labels.VALUE: args[0]}
        else:
            kwargs = _combine_args_with_kwargs(s.outputs.keys(), *args, **kwargs)

        s.set_graph_outputs(**kwargs)
        for o in kwargs:
            if ot := s.outputs.get(o):
                g.annotate_output(o, ot)

        super().__init__(g.output.idx, g)


class _LoopOutput(Output):
    def __init__(self, loopout: ValueSource) -> None:
        super().__init__(value=loopout)


class Continue(_LoopOutput):
    def __init__(self, loopout: ValueSource) -> None:
        super().__init__(Tag(Labels.CONTINUE, loopout))


class Break(_LoopOutput):
    def __init__(self, loopout: ValueSource) -> None:
        super().__init__(Tag(Labels.BREAK, loopout))


def _vals_to_types(
    d: dict[str, Optional[Type["TierkreisValue"]]],
) -> dict[str, Optional["TierkreisType"]]:
    return map_vals(
        d, lambda arg: None if (arg is Any or arg is None) else tkvalue_to_tktype(arg)
    )


def _convert_input(t: Optional[Type["Input[TierkreisValue]"]]) -> Optional[Type]:
    if t is not None:
        if not (get_origin(t) == Input or t == Input):
            raise TypeError(
                "Graph builder function arguments must be ``Input``"
                " with an optional ``TierkreisValue`` argument."
                "For example Input[IntValue]"
            )

        args = get_args(t)
        if args:
            return args[0]
    return None


_RETURN_TYPE_ERROR = TypeError(
    "Graph builder function must be annotated with return type `Output`."
    " Optionally annotate with "
    "`Output[TierkreisValue]` for a single return from port 'value',"
    " or with a `UnpackRow` specifying the Row type of the ouput."
)


def _convert_return(ret: Type) -> dict[str, Optional["TierkreisType"]]:
    if not (get_origin(ret) == Output or ret == Output):
        raise _RETURN_TYPE_ERROR
    args = get_args(ret)
    if not args:
        return {}
    ret = args[0]
    subc = cast(Type, get_origin(ret) or ret)
    if issubclass(subc, UnpackRow):
        return _vals_to_types(get_type_hints(ret))
    if issubclass(subc, TierkreisValue):
        return _vals_to_types({Labels.VALUE: ret})

    raise _RETURN_TYPE_ERROR


class Input(Generic[TKVal1], NodePort):
    def __init__(self, port: PortID) -> None:
        np = current_builder().graph.input[port]
        super().__init__(np.node_ref, np.port)


_GraphDef = Callable[..., Output]


def _get_edge_annotations(
    f: _GraphDef,
) -> tuple[
    dict[str, Optional["TierkreisType"]],
    dict[str, Optional["TierkreisType"]],
]:
    f_sig = inspect.signature(f)
    inps = map_vals(
        inspect.signature(f).parameters,
        lambda x: None if x.annotation is inspect.Signature.empty else x.annotation,
    )
    out_types = (
        {}
        if f_sig.return_annotation is inspect.Signature.empty
        else _convert_return(f_sig.return_annotation)
    )
    in_types = _vals_to_types(map_vals(inps, _convert_input))
    return in_types, out_types


class CapturedError(Exception):
    pass


@dataclass(frozen=True)
class LazyGraph:
    _builder: Callable[[], TierkreisGraph]

    @cached_property
    def graph(self) -> TierkreisGraph:
        return self._builder()

    def __call__(self, *args: ValueSource, **kwargs: ValueSource) -> NodeRef:
        return self.graph(*args, **kwargs)

    def _repr_svg_(self) -> str:
        return getattr(self.graph, "_repr_svg_")()


_GraphDecoratorType = Callable[[_GraphDef], TierkreisGraph]
_LazyGraphDecoratorType = Callable[[_GraphDef], LazyGraph]


def lazy_graph(
    name: Optional[str] = None,
    type_check_sig: Optional[Signature] = None,
    debug: bool = True,
) -> _LazyGraphDecoratorType:
    def decorator_graph(f: _GraphDef) -> LazyGraph:
        in_types, out_types = _get_edge_annotations(f)
        # Use getattr as _GraphDef doesn't specify that it has a __name__
        graph_name = name or getattr(f, "__name__")
        # graph_name = name or f.__name__  # type: ignore  # the alternative

        def wrapper() -> TierkreisGraph:
            gb = GraphBuilder(in_types, out_types, name=graph_name, debug=debug)

            with gb as sub_build:
                f(*(Input(port) for port in in_types))

            return (
                sub_build.graph
                if type_check_sig is None
                else sub_build.type_check(type_check_sig)
            )

        return LazyGraph(wrapper)

    return decorator_graph


def graph(
    name: Optional[str] = None,
    type_check_sig: Optional[Signature] = None,
    debug: bool = True,
) -> _GraphDecoratorType:
    dec = lazy_graph(name, type_check_sig, debug)

    def decorator_graph(f: _GraphDef) -> TierkreisGraph:
        return dec(f).graph

    return decorator_graph


class Thunk(_CallAddNode, PortFunc):
    graph_src: ValueSource

    def __init__(
        self, graph_src: ValueSource, input_order: list[str], output_order: list[str]
    ):
        self.graph_src = graph_src
        super().__init__(FunctionNode(FunctionName("eval")), input_order, output_order)

    def __call__(self, *args: ValueSource, **kwargs: ValueSource) -> NodeRef:
        assert "thunk" not in kwargs
        return super().__call__(*args, thunk=self.graph_src, **kwargs)

    def copyable(self) -> "Thunk":
        self.graph_src = Copyable(self.graph_src)
        return self

    def resolve(self) -> NodePort:
        """Return a NodePort, creating nodes as necessary in source graph."""
        return _resolve(self.graph_src)

    def source_graph(self) -> TierkreisGraph:
        """The graph the port belongs to (or will)."""
        return _source_graph(self.graph_src)


def closure(
    name: Optional[str] = None, debug: bool = True
) -> Callable[[_GraphDef], Thunk]:
    def decorator_graph(f: _GraphDef) -> Thunk:
        in_types, out_types = _get_edge_annotations(f)
        # Use getattr as _GraphDef doesn't specify that it has a __name__
        graph_name = name or getattr(f, "__name__")
        # graph_name = name or f.__name__  # type: ignore  # the alternative

        gb = CaptureBuilder(in_types, out_types, name=graph_name, debug=debug)

        with gb as sub_build:
            f(*(Input(port) for port in in_types))
        return Thunk(
            _partial_thunk(sub_build, sub_build.captured),
            list(in_types.keys()),
            list(out_types.keys()),
        )

    return decorator_graph


def loop(name: Optional[str] = None, debug: bool = True):
    def decorator_graph(f: _GraphDef):
        in_types, out_types = _get_edge_annotations(f)
        if len(in_types) != 1:
            raise ValueError("Loop body graph can only have one input.")

        # the loop node has input port "value"
        # allow body definitions to use any input names and map it
        in_types = {Labels.VALUE: in_types.popitem()[1]}
        graph_name = name or getattr(f, "__name__")

        def wrapper(initial: ValueSource) -> NodePort:
            gb = CaptureBuilder(in_types, out_types, name=graph_name, debug=debug)
            with gb as sub_build:
                f(Input(Labels.VALUE))

            loop_node = current_builder().add_node_to_graph(
                FunctionNode(FunctionName("loop")),
                body=_partial_thunk(sub_build, sub_build.captured),
                value=initial,
            )
            return loop_node[Labels.VALUE]

        return wrapper

    return decorator_graph


class If(CaptureBuilder):
    def __init__(self) -> None:
        super().__init__()
        self.graph.name = "if"
        Match._add_handler("if", self)


class Else(CaptureBuilder):
    def __init__(self) -> None:
        super().__init__()
        self.graph.name = "else"
        Match._add_handler("else", self)


class DuplicateBlock(Exception):
    """Duplicate blocks, avoid multiple If/Else or duplicate Case blocks."""

    block_name: str


class IntermediateNodes(Exception):
    """Adding of nodes not valid in _CaseScope outside child blocks."""


class _CaseScope(GraphBuilder, ABC):
    __variant_handlers: ContextVar[dict[str, CaptureBuilder]] = ContextVar(
        "variant_handlers"
    )

    _token: Token[dict[str, CaptureBuilder]]
    nref: NodeRef

    @classmethod
    def _get_handlers(cls) -> dict[str, CaptureBuilder]:
        return cls.__variant_handlers.get()

    def __enter__(self) -> "_CaseScope":
        super().__enter__()
        self._token = self.__variant_handlers.set({})
        return self

    @classmethod
    def _add_handler(cls, tag: str, handler: CaptureBuilder):
        handlers = cls.__variant_handlers.get()
        if tag in handlers:
            raise DuplicateBlock(tag)
        handlers[tag] = handler

    @abstractmethod
    def _get_thunk(self) -> NodePort: ...

    @abstractmethod
    def _intermediate_error(self) -> IntermediateNodes: ...

    def __exit__(
        self,
        __exc_type: Optional[Type[BaseException]],
        __exc_value: Optional[BaseException],
        __traceback: Optional[TracebackType],
    ) -> None:
        super().__exit__(__exc_type, __exc_value, __traceback)

        if not any((__exc_type, __exc_value, __traceback)):
            # do not perform graph building if error

            if self.graph.n_nodes != 2:
                raise self._intermediate_error()
            handlers = self.__variant_handlers.get()
            for h in handlers.values():
                current_builder().debug_info.update(h.debug_info)

            thunk_port = self._get_thunk()
            partial_inps = _combine_captures(list(handlers.values()))
            thunk_port = _partial_thunk(thunk_port, partial_inps)

            self.nref = current_builder().add_node_to_graph(
                FunctionNode(FunctionName("eval")),
                thunk=thunk_port,
            )
        self.__variant_handlers.reset(self._token)


class Match(_CaseScope):
    variant_value: ValueSource

    def __init__(self, variant_value: ValueSource):
        super().__init__()
        self.variant_value = variant_value

    def _get_thunk(self) -> NodePort:
        handlers = _CaseScope._get_handlers()
        bg = current_builder()
        return bg.add_node_to_graph(
            MatchNode(),
            variant_value=self.variant_value,
            **map_vals(handlers, lambda x: Const(x.graph)),
        )[Labels.THUNK]

    def _intermediate_error(self) -> IntermediateNodes:
        return IntermediateNodes(
            "Graph building operations outside Case blocks not allowed in Match block."
        )


class Case(CaptureBuilder):
    var_value: NodePort

    def __init__(self, tag: str) -> None:
        super().__init__()
        self.graph.name = tag
        Match._add_handler(tag, self)

        self.var_value = self.graph.input[Labels.VALUE]


class IfElse(_CaseScope):
    predicate: ValueSource

    def __init__(self, predicate: ValueSource):
        super().__init__()
        self.predicate = predicate

    def _get_thunk(self) -> NodePort:
        handlers = _CaseScope._get_handlers()
        assert len(handlers) == 2
        bg = current_builder()
        return bg.add_node_to_graph(
            FunctionNode(FunctionName("switch")),
            pred=self.predicate,
            if_true=Const(handlers["if"].graph),
            if_false=Const(handlers["else"].graph),
        )[Labels.VALUE]

    def _intermediate_error(self) -> IntermediateNodes:
        return IntermediateNodes(
            "Graph building operations outside"
            " If/Else blocks not allowed in IfElse block."
        )


def _combine_captures(thunks: list[CaptureBuilder]) -> dict[ValueSource, PortID]:
    """Given a list of thunks with captured inputs, remove their existing input
    names, common up any shared sources (up to == equality), and add all those
    inputs to all graphs, discarding any unused."""
    old_names = frozenset([v for bg in thunks for v in bg.captured.values()])
    # Generate a sparsely numbered list of names that won't conflict:
    names = filter(lambda x: x not in old_names, map(_capture_label, count()))
    # Assign new names in deterministic order of thunks and captures in each
    combined_inputs: dict[ValueSource, str] = {
        vs: next(names) for bg in thunks for vs in bg.captured
    }  # Duplicate ValueSources overwrite earlier entries, making even sparser.
    for vs, nam in combined_inputs.items():
        for bg in thunks:
            if in_name := bg.captured.get(vs):
                in_edge = bg.graph.out_edge_from_port(bg.graph.input[in_name])
                assert in_edge is not None
                bg.graph._graph.remove_edge(*in_edge.to_edge_handle())
                bg.graph.add_edge(bg.graph.input[nam], in_edge.target)
            else:
                bg.graph.discard(bg.graph.input[nam])

    return combined_inputs


def _combine_args_with_kwargs(
    expected: Iterable[str], *args: ValueSource, **kwargs: ValueSource
) -> dict[str, ValueSource]:
    pos_args = dict(zip(expected, args))
    if ints := set(kwargs).intersection(pos_args):
        raise RuntimeError(f"Arguments specified positionally and in keywords: {ints}")
    kwargs.update(pos_args)
    return kwargs


class Copyable(PortFunc):
    """Inserts a copy of an underlying ValueSource whenever an outgoing edge is added
    unless there are *no* existing uses of the underlying ValueSource
    (=> the wire is routed from the underlying ValueSource)
    """

    np: NodePort

    def __init__(self, src: ValueSource):
        builder = current_builder()
        while isinstance(src, Copyable) and _source_graph(src) == builder.graph:
            # Simplify if we call Copyable on a ValueSource twice
            self = src
        self.np = _resolve(builder.capture(src, allow_existing=True))

    def resolve(self) -> NodePort:
        g = self.source_graph()
        existing_edge = g.out_edge_from_port(self.np)
        if existing_edge is None:
            return self.np
        g._graph.remove_edge(*existing_edge.to_edge_handle())
        c1, c2 = g.copy_value(self.np)
        g.add_edge(c1, existing_edge.target, existing_edge.type_)
        return c2

    def source_graph(self) -> "TierkreisGraph":
        return self.np.node_ref.graph


@dataclass(frozen=True)
class Unpack(StablePortFunc):
    src: Union[NodePort, StablePortFunc]
    field_name: str

    def resolve(self) -> NodePort:
        np = _resolve(self.src)
        g = np.node_ref.graph
        existing_edge = g.out_edge_from_port(np)
        if existing_edge is not None:
            target_node = g[existing_edge.target.node_ref]
            if not target_node.is_unpack_node():
                raise ValueError(
                    "Cannot unpack port wired to something "
                    f" other than an unpack_struct: {target_node}"
                )
            unpack_node = existing_edge.target.node_ref
        else:
            unpack_node = g.add_func("unpack_struct", struct=np)
        return unpack_node[self.field_name]

    def source_graph(self) -> "TierkreisGraph":
        return _source_graph(self.src)


def __unpack_nodeport(self: NodePort, field_name: str) -> Unpack:
    return Unpack(self, field_name)


NodePort.__getitem__ = __unpack_nodeport  # type: ignore


def _arg_str(args: Dict[str, "TierkreisType"], order: Iterable[str]) -> str:
    return ", ".join(f"{chalk.yellow(port)}: {args[port]}" for port in order)


def _func_sig(name: str, func: FunctionDeclaration):
    graph_type = cast("GraphType", TypeScheme.from_proto(func.type_scheme).body)
    irest = graph_type.inputs.rest
    orest = graph_type.outputs.rest
    irest = f", {chalk.yellow('#')}: {irest}" if irest else ""
    orest = f", {chalk.yellow('#')}: {orest}" if orest else ""
    return (
        f"{chalk.bold.blue(name)}"
        f"({_arg_str(graph_type.inputs.content, func.input_order)}{irest})"
        f" -> ({_arg_str(graph_type.outputs.content, func.output_order)}{orest})"
    )


class Function(_CallAddNode):
    f: FunctionDeclaration
    name: FunctionName

    def __init__(self, f: FunctionDeclaration, name: FunctionName):
        self.f = f
        self.name = name
        super().__init__(FunctionNode(name), f.input_order, f.output_order)

    def signature_string(self) -> str:
        return _func_sig(self.name.name, self.f)

    def to_graph(
        self,
        input_map: Optional[dict[str, str]] = None,
        output_map: Optional[dict[str, str]] = None,
    ) -> TierkreisGraph:
        return graph_from_func(self.name, self.f, input_map, output_map)

    def __str__(self) -> str:
        return self.signature_string()


class Namespace(Mapping[str, "Namespace"]):
    @overload
    def __init__(self, args: Signature, /): ...

    @overload
    def __init__(self, args: SigNamespace, *, _prefix: list[str]):
        """For internal use only."""
        ...

    def __init__(self, args: Union[Signature, SigNamespace], **kwargs: list[str]):
        if isinstance(args, Signature):
            namespace = args.root
            prefix = []
        else:
            namespace = args
            prefix = kwargs["_prefix"]
        for name, ndef in namespace.functions.items():
            # allows auto completion of function names in Jupyter environment
            setattr(self, name, Function(ndef, FunctionName(name, prefix)))
        self.__subspaces: Mapping[str, "Namespace"] = {
            name: Namespace(ns, _prefix=prefix + [name])
            for name, ns in namespace.subspaces.items()
        }

    def __getattr__(
        self, attr_name: str
    ) -> Function:  # This is here just to placate mypy
        raise AttributeError(attr_name)

    def __getitem__(self, __k: str) -> "Namespace":
        return self.__subspaces[__k]

    def __len__(self) -> int:
        return self.__subspaces.__len__()

    def __iter__(self) -> Iterator[str]:
        return self.__subspaces.__iter__()


def _rename_ports(
    thunk: ValueSource, map_old_to_new: dict[str, str], is_inputs: bool
) -> NodePort:
    bg = current_builder()
    rename_g = Const(rename_ports_graph(map_old_to_new))
    f, s = (rename_g, thunk) if is_inputs else (thunk, rename_g)
    seq = bg.add_node_to_graph(
        FunctionNode(FunctionName("sequence")), first=f, second=s
    )

    return seq["sequenced"]


def RenameInputs(thunk: ValueSource, rename_map: dict[str, str]) -> NodePort:
    return _rename_ports(thunk, {v: k for k, v in rename_map.items()}, True)


def RenameOuputs(thunk: ValueSource, rename_map: dict[str, str]) -> NodePort:
    return _rename_ports(thunk, rename_map, False)


def _build_box(
    g: TierkreisGraph, __tk_loc: Location, *args: ValueSource, **kwargs: ValueSource
) -> NodeRef:
    bx = _CallAddNode(
        BoxNode(g, __tk_loc),
        g.input_order,
        g.output_order,
    )
    return bx(*args, **kwargs)


def __graph_call__(
    self: TierkreisGraph, *args: ValueSource, **kwargs: ValueSource
) -> NodeRef:
    return _build_box(self, Location([]), *args, **kwargs)


def UnpackTuple(src: ValueSource, first_n: int) -> Iterator[NodePort]:
    """Unpack a struct containing python tuple fields, up to the `first_n` fields."""
    unpack_node = current_builder().add_node_to_graph(
        FunctionNode(FunctionName("unpack_struct")),
        struct=src,
    )

    return (unpack_node[TupleLabel.index_label(i)] for i in range(first_n))


def MakeTuple(*args: ValueSource) -> NodePort:
    """Make a struct from some values, treating them like a python tuple, with
    member position indices used to generate the field labels."""
    make_node = current_builder().add_node_to_graph(
        FunctionNode(FunctionName("make_struct")),
        **{TupleLabel.index_label(i): v for i, v in enumerate(args)},
    )
    return make_node["struct"]


# mypy is not happy with assignment to a method
setattr(TierkreisGraph, "__call__", __graph_call__)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: v1alpha/runtime.proto
# plugin: python-betterproto
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Dict,
    List,
    Optional,
)

import betterproto
import grpclib
from betterproto.grpc.grpclib_server import ServiceBase

from .. import (
    graph as _graph__,
    signature as _signature__,
)


if TYPE_CHECKING:
    import grpclib.server
    from betterproto.grpc.grpclib_client import MetadataLike
    from grpclib.metadata import Deadline


@dataclass(eq=False, repr=False)
class Callback(betterproto.Message):
    uri: str = betterproto.string_field(1)
    loc: "_graph__.Location" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class RunTaskRequest(betterproto.Message):
    graph: "_graph__.Graph" = betterproto.message_field(1)
    inputs: "_graph__.StructValue" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class RunTaskResponse(betterproto.Message):
    task_id: str = betterproto.string_field(1, group="result")
    type_errors: "_signature__.TypeErrors" = betterproto.message_field(
        2, group="result"
    )


@dataclass(eq=False, repr=False)
class ListTasksRequest(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class ListTasksResponse(betterproto.Message):
    tasks: List["Task"] = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class DeleteTaskRequest(betterproto.Message):
    id: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class DeleteTaskResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class GetTaskRequest(betterproto.Message):
    id: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class GetTaskResponse(betterproto.Message):
    task: "Task" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class AwaitTaskRequest(betterproto.Message):
    id: str = betterproto.string_field(1)
    timeout: int = betterproto.uint64_field(2)


@dataclass(eq=False, repr=False)
class AwaitTaskResponse(betterproto.Message):
    task: "Task" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class Task(betterproto.Message):
    id: str = betterproto.string_field(1)
    success: "_graph__.StructValue" = betterproto.message_field(2, group="status")
    error: str = betterproto.string_field(3, group="status")


@dataclass(eq=False, repr=False)
class RunGraphRequest(betterproto.Message):
    graph: "_graph__.Graph" = betterproto.message_field(1)
    inputs: "_graph__.StructValue" = betterproto.message_field(2)
    type_check: bool = betterproto.bool_field(3)
    loc: "_graph__.Location" = betterproto.message_field(4)
    escape: "Callback" = betterproto.message_field(5)


@dataclass(eq=False, repr=False)
class RunGraphResponse(betterproto.Message):
    success: "_graph__.StructValue" = betterproto.message_field(1, group="result")
    error: str = betterproto.string_field(2, group="result")
    type_errors: "_signature__.TypeErrors" = betterproto.message_field(
        3, group="result"
    )


class RuntimeStub(betterproto.ServiceStub):
    async def list_tasks(
        self,
        list_tasks_request: "ListTasksRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "ListTasksResponse":
        return await self._unary_unary(
            "/tierkreis.v1alpha.runtime.Runtime/ListTasks",
            list_tasks_request,
            ListTasksResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def run_task(
        self,
        run_task_request: "RunTaskRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "RunTaskResponse":
        return await self._unary_unary(
            "/tierkreis.v1alpha.runtime.Runtime/RunTask",
            run_task_request,
            RunTaskResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def get_task(
        self,
        get_task_request: "GetTaskRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "GetTaskResponse":
        return await self._unary_unary(
            "/tierkreis.v1alpha.runtime.Runtime/GetTask",
            get_task_request,
            GetTaskResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def await_task(
        self,
        await_task_request: "AwaitTaskRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "AwaitTaskResponse":
        return await self._unary_unary(
            "/tierkreis.v1alpha.runtime.Runtime/AwaitTask",
            await_task_request,
            AwaitTaskResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def delete_task(
        self,
        delete_task_request: "DeleteTaskRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "DeleteTaskResponse":
        return await self._unary_unary(
            "/tierkreis.v1alpha.runtime.Runtime/DeleteTask",
            delete_task_request,
            DeleteTaskResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def run_graph(
        self,
        run_graph_request: "RunGraphRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "RunGraphResponse":
        return await self._unary_unary(
            "/tierkreis.v1alpha.runtime.Runtime/RunGraph",
            run_graph_request,
            RunGraphResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )


class RuntimeBase(ServiceBase):
    async def list_tasks(
        self, list_tasks_request: "ListTasksRequest"
    ) -> "ListTasksResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def run_task(self, run_task_request: "RunTaskRequest") -> "RunTaskResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def get_task(self, get_task_request: "GetTaskRequest") -> "GetTaskResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def await_task(
        self, await_task_request: "AwaitTaskRequest"
    ) -> "AwaitTaskResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def delete_task(
        self, delete_task_request: "DeleteTaskRequest"
    ) -> "DeleteTaskResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def run_graph(
        self, run_graph_request: "RunGraphRequest"
    ) -> "RunGraphResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def __rpc_list_tasks(
        self, stream: "grpclib.server.Stream[ListTasksRequest, ListTasksResponse]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.list_tasks(request)
        await stream.send_message(response)

    async def __rpc_run_task(
        self, stream: "grpclib.server.Stream[RunTaskRequest, RunTaskResponse]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.run_task(request)
        await stream.send_message(response)

    async def __rpc_get_task(
        self, stream: "grpclib.server.Stream[GetTaskRequest, GetTaskResponse]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.get_task(request)
        await stream.send_message(response)

    async def __rpc_await_task(
        self, stream: "grpclib.server.Stream[AwaitTaskRequest, AwaitTaskResponse]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.await_task(request)
        await stream.send_message(response)

    async def __rpc_delete_task(
        self, stream: "grpclib.server.Stream[DeleteTaskRequest, DeleteTaskResponse]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.delete_task(request)
        await stream.send_message(response)

    async def __rpc_run_graph(
        self, stream: "grpclib.server.Stream[RunGraphRequest, RunGraphResponse]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.run_graph(request)
        await stream.send_message(response)

    def __mapping__(self) -> Dict[str, grpclib.const.Handler]:
        return {
            "/tierkreis.v1alpha.runtime.Runtime/ListTasks": grpclib.const.Handler(
                self.__rpc_list_tasks,
                grpclib.const.Cardinality.UNARY_UNARY,
                ListTasksRequest,
                ListTasksResponse,
            ),
            "/tierkreis.v1alpha.runtime.Runtime/RunTask": grpclib.const.Handler(
                self.__rpc_run_task,
                grpclib.const.Cardinality.UNARY_UNARY,
                RunTaskRequest,
                RunTaskResponse,
            ),
            "/tierkreis.v1alpha.runtime.Runtime/GetTask": grpclib.const.Handler(
                self.__rpc_get_task,
                grpclib.const.Cardinality.UNARY_UNARY,
                GetTaskRequest,
                GetTaskResponse,
            ),
            "/tierkreis.v1alpha.runtime.Runtime/AwaitTask": grpclib.const.Handler(
                self.__rpc_await_task,
                grpclib.const.Cardinality.UNARY_UNARY,
                AwaitTaskRequest,
                AwaitTaskResponse,
            ),
            "/tierkreis.v1alpha.runtime.Runtime/DeleteTask": grpclib.const.Handler(
                self.__rpc_delete_task,
                grpclib.const.Cardinality.UNARY_UNARY,
                DeleteTaskRequest,
                DeleteTaskResponse,
            ),
            "/tierkreis.v1alpha.runtime.Runtime/RunGraph": grpclib.const.Handler(
                self.__rpc_run_graph,
                grpclib.const.Cardinality.UNARY_UNARY,
                RunGraphRequest,
                RunGraphResponse,
            ),
        }



from typing import cast

import pytest
from pytket.backends.backendresult import BackendResult

from tierkreis import TierkreisGraph
from tierkreis.common_types.circuit import register_pytket_types
from tierkreis.core.values import StructValue, VecValue
from tierkreis.pyruntime import PyRuntime

register_pytket_types()


@pytest.fixture
def bell_circuit() -> str:
    return (
        r'{"bits": [["c", [0]], ["c", [1]]], "commands": [{"args": [["q", [0]]], "op": '
        r'{"type": "H"}}, {"args": [["q", [0]], ["q", [1]]], "op": {"type": "CX"}}, '
        r'{"args": [["q", [0]], ["c", [0]]], "op": {"type": "Measure"}}, {"args": '
        r'[["q", [1]], ["c", [1]]], "op": {"type": "Measure"}}], '
        r'"implicit_permutation": [[["q", [0]], ["q", [0]]], [["q", [1]], ["q", '
        r'[1]]]], "phase": "0.0", "qubits": [["q", [0]], ["q", [1]]]}'
    )


@pytest.fixture
def pytket_pyruntime():
    import pytket_worker.main  # type: ignore

    return PyRuntime([pytket_worker.main.root])


@pytest.mark.asyncio
@pytest.mark.pytket
async def test_compile_circuit(bell_circuit: str, pytket_pyruntime) -> None:
    tg = TierkreisGraph()
    load_node = tg.add_func("pytket::load_circuit_json", json_str=tg.input["input"])
    compile_node = tg.add_func(
        "pytket::compile_circuits",
        circuits=tg.make_vec([load_node["value"]]),
        pass_name=tg.add_const("FullPeepholeOptimise"),
    )
    tg.set_outputs(out=compile_node)

    outs = await pytket_pyruntime.run_graph(tg, input=bell_circuit)

    assert isinstance(cast(VecValue, outs["out"]).values[0], StructValue)


@pytest.mark.asyncio
@pytest.mark.pytket
async def test_execute_circuit(bell_circuit: str, pytket_pyruntime) -> None:
    tg = TierkreisGraph()
    vec_elems = tg.vec_last_n_elems(tg.input["circuits"], 2)
    load_node1 = tg.add_func("pytket::load_circuit_json", json_str=vec_elems[0])
    load_node2 = tg.add_func("pytket::load_circuit_json", json_str=vec_elems[1])
    execute_node = tg.add_func(
        "pytket::execute_circuits",
        circuits=tg.make_vec([load_node1, load_node2]),
        shots=tg.input["shots"],
        backend_name=tg.add_const("AerBackend"),
    )
    tg.set_outputs(out=execute_node)

    shots = [10, 20]
    outputs = cast(
        VecValue,
        (
            await pytket_pyruntime.run_graph(
                tg,
                circuits=[bell_circuit, bell_circuit],
                shots=shots,
            )
        )["out"],
    )

    for count, res in zip(shots, outputs.values):
        assert isinstance(res, StructValue)

        pytket_res = res.to_python(BackendResult)

        assert sum(pytket_res.get_counts().values()) == count



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: v1alpha1/worker.proto
# plugin: python-betterproto
# This file has been @generated

from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Dict,
    Optional,
)

import betterproto
import grpclib
from betterproto.grpc.grpclib_server import ServiceBase

from .. import (
    graph as _graph__,
    runtime as _runtime__,
)


if TYPE_CHECKING:
    import grpclib.server
    from betterproto.grpc.grpclib_client import MetadataLike
    from grpclib.metadata import Deadline


@dataclass(eq=False, repr=False)
class RunFunctionRequest(betterproto.Message):
    function: "_graph__.FunctionName" = betterproto.message_field(1)
    inputs: "_graph__.StructValue" = betterproto.message_field(2)
    loc: "_graph__.Location" = betterproto.message_field(3)
    callback: "_runtime__.Callback" = betterproto.message_field(4)


@dataclass(eq=False, repr=False)
class RunFunctionResponse(betterproto.Message):
    outputs: "_graph__.StructValue" = betterproto.message_field(1)


class WorkerStub(betterproto.ServiceStub):
    async def run_function(
        self,
        run_function_request: "RunFunctionRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "RunFunctionResponse":
        return await self._unary_unary(
            "/tierkreis.v1alpha1.worker.Worker/RunFunction",
            run_function_request,
            RunFunctionResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )


class WorkerBase(ServiceBase):

    async def run_function(
        self, run_function_request: "RunFunctionRequest"
    ) -> "RunFunctionResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def __rpc_run_function(
        self, stream: "grpclib.server.Stream[RunFunctionRequest, RunFunctionResponse]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.run_function(request)
        await stream.send_message(response)

    def __mapping__(self) -> Dict[str, grpclib.const.Handler]:
        return {
            "/tierkreis.v1alpha1.worker.Worker/RunFunction": grpclib.const.Handler(
                self.__rpc_run_function,
                grpclib.const.Cardinality.UNARY_UNARY,
                RunFunctionRequest,
                RunFunctionResponse,
            ),
        }
